/*

    MidiIn -- C/python module for parsing MIDI input

    Copyright (C) 1999  Eric S. Tiedemann

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact: Eric S. Tiedemann, est@hyperreal.org

*/

#include <stdlib.h>

#ifdef PYTHON
#include <python1.5/Python.h>
#endif

static const int EOX = 0xf7;
static const int SEX = 0xf0;

#define MAX_COMPRESSED_CONTROLLER 95

struct MidiIn
{
  unsigned char cp[MAX_COMPRESSED_CONTROLLER + 1];
  unsigned char cval[MAX_COMPRESSED_CONTROLLER + 1];
  int status; 
  int status3;
  int thirdp;
  int second;
  int (*cb)(void *, const char *, int, int, int);
  void *cb_data;
  int compressionp;
};

  
struct MidiIn *
MidiIn_New()
{
  struct MidiIn *m = (struct MidiIn *) calloc(1, sizeof(struct MidiIn));

  return m;
}

static int
channel(int b)
{
    return (b & 0xF) + 1;
}

static const char *
status2name(int s)
{
  switch (s < 0xF0 ? (s & 0xF0) : s)
    {
    case 0x80: return("note-off");
    case 0x90: return("note-on");
    case 0xa0: return("poly-key-pressure");
    case 0xb0: return("control-change");
    case 0xc0: return("program-change");
    case 0xd0: return("channel-pressure");
    case 0xe0: return("pitch-bend");
    case 0xf0: return("system-exclusive");
    case 0xf1: return("MIDI-Time-Code-quarter-frame");
    case 0xf2: return("song-position-pointer");
    case 0xf3: return("song-select");
    case 0xf6: return("tune-request");
    case 0xf7: return("end-of-exclusive");
    case 0xf8: return("timing-clock");
    case 0xfa: return("start");
    case 0xfb: return("continue");
    case 0xfc: return("stop");
    case 0xfe: return("active-sensing");
    case 0xff: return("system-reset");
    default: return "???";
    }
}

static const char *
channel_mode_message(int i)
{
  switch (i)
    {
    case 120: return("all-sound-off");
    case 121: return("reset-all-controllers");
    case 122: return("local-control");
    case 123: return("all-notes-off");
    case 124: return("omni-off");
    case 125: return("omni-on");
    case 126: return("mono-on");
    case 127: return("poly-on");
    default: return "???";
    }
}

static int
invoke_callback(struct MidiIn *m, int e, int v, int w)
{
  /*printf("%02x %d %d\n", e, v, w);*/

  if (m->cb)
    {
      if ((e & 0xF0) == 0xB0)
	{
	  /* handle special values of v for control change*/
	  if (v > 119 && v < 128)
	    return m->cb(m->cb_data, channel_mode_message(v), 0, 0, 0);
	  else
	    {
	      if (m->compressionp 
		  && (e & 0xF0) == 0xB0 
		  && channel(e) == 5 
		  && v <= MAX_COMPRESSED_CONTROLLER)
		{
		  m->cp[v] = 1;
		  m->cval[v] = w;
		  /* invalidate compressed lsbs when receiving an msb */
		  if (v < 32 && m->cp[v + 32])
		    m->cp[v + 32] = 0;
		}
	      else
		return m->cb(m->cb_data, status2name(e), channel(e), v, w);
	    }
	}
      else if (e == SEX)
	return m->cb(m->cb_data, status2name(e), 0, 0, 0);
      else if ((e & 0xF0) == 0xE0)
	return m->cb(m->cb_data, status2name(e), channel(e), v + w * 128, 0);
      else if (e == 0xF2)
	return m->cb(m->cb_data, status2name(e), v + w * 128, 0, 0);
      else if (e < 0xF0)
	return m->cb(m->cb_data, status2name(e), channel(e), v, w);
      else
	return m->cb(m->cb_data, status2name(e), v, w, 0);
    }
}

static int
real_time(int b)
{
  switch (b)
  {
    case 0xf8:
    case 0xfa:
    case 0xfb: 
    case 0xfc: 
    case 0xfe: 
    case 0xff: return 1;
    default: return 0;
  }
}

static int
status_byte(int b)
{
  return b & 0x80;
}

static void
sysexb(struct MidiIn *m, int b)
{
}

static int
sysexend(struct MidiIn *m)
{
  return 1;
}

int
MidiIn_InByte(struct MidiIn *m, int b)
{ 
  /*printf("%d\n", b);*/
  int rc = 1;

  if (status_byte(b))
    {
      if (real_time(b))
	rc = invoke_callback(m, b, 0, 0);
      else if (b == EOX)
	{
	  if (m->status == SEX)
	    {
	      rc = sysexend(m);
	      m->status = 0;
	    }
	  else
	    /* complain? */
	    ;			
	}
      else
	{
	  if (m->status == SEX)
	    rc = invoke_callback(m, SEX, 0, 0);
	  m->status = b;
	  if (b == 0xF6)
	    {
	      if (!invoke_callback(m, b, 0, 0))
		rc = 0;
	    }
	}
    }
  else				/* data byte */
    {
      if (m->status == SEX)
	sysexb(m, b);
      else if (m->thirdp)
	{
	  if (m->status3)
	    {
	      rc = invoke_callback(m, m->status3, m->second, b);
	      m->status3 = 0;
	    }
	  else
	    rc = invoke_callback(m, m->status, m->second, b);
	  m->thirdp = 0;
	}
      else
	{
	  if (m->status)
	    {
	      if (m->status < 0xC0)
		{
		  m->thirdp = 1;
		  m->second = b;
		}
	      else
		{
		  if (m->status < 0xE0)
		    rc = invoke_callback(m, m->status, b, 0);
		  else
		    {
		      if (m->status < 0xF0)
			{
			  m->thirdp = 1;
			  m->second = b;
			}
		      else
			{
			  if (m->status == 0xF2)
			    {
			      m->thirdp = 1;
			      m->second = b;
			      m->status3 = m->status;
			      m->status = 0;
			    }
			  else
			    {
			      if (m->status == 0xF3 || m->status == 0xF1)
				{
				  rc = invoke_callback(m, m->status, b, 0);
				  m->status = 0;
				}
			      else
				{
				  /* ignore strange status byte */
				  m->status = 0; 
				}
			    }
			}
		    }
		}
	    }
	  else
	    ; /* ignore data byte with no status on record */
	}
    }

  return rc;
}

int
MidiIn_InBytes(struct MidiIn *m, unsigned char *bs, size_t n)
{ 
  size_t i;
  int rc = 1;

  m->compressionp = 1;
  for (i = 0; i <= MAX_COMPRESSED_CONTROLLER; i++)
    m->cp[i] = 0;
  for (i = 0; i < n; i++)
    if (!MidiIn_InByte(m, bs[i]))
      rc = 0;
  for (i = 0; i <= MAX_COMPRESSED_CONTROLLER; i++)
    if(m->cp[i])
      {
	if (!(m->cb(m->cb_data, "control-change", 5, i, m->cval[i])))
	  rc = 0;
      }

  return rc;
}

#ifdef PYTHON

typedef struct
{
  PyObject_HEAD
  struct MidiIn *o;
} PyMidiIn;

#define PyMidiIn_Check(o) ((o)->ob_type == &PyMidiIn_Type)

static PyObject *
PyMidiIn_inbytes(PyObject *self, PyObject *args)
{
  unsigned char *s;
  int n;

  if (!PyArg_Parse(args, "s#;midiin.inbytes", &s, &n))
    return NULL;

  if (!MidiIn_InBytes(((PyMidiIn *) self)->o, s, n))
    return 0;

  Py_INCREF(Py_None);
  return Py_None;
}

static struct PyMethodDef PyMidiIn_methods[] = {
  {"inbytes", PyMidiIn_inbytes},
  {NULL, NULL}
};

static PyObject *
PyMidiIn_getattr(PyObject *self, char *name)
{
  return Py_FindMethod(PyMidiIn_methods, self, name);
}

static void
PyMidiIn_del(PyMidiIn *o)
{
  return;
}

static PyTypeObject PyMidiIn_Type =
{
  PyObject_HEAD_INIT(&PyType_Type)
  0,
  "MidiIn",
  sizeof(PyMidiIn),
  0,

  (destructor) PyMidiIn_del,
  0,
  PyMidiIn_getattr,
  0,
  0,
  0,

  0,
  0,
  0,

  0,
  0,
  0,
};

static int
callcb(void *d, const char *s, int c, int v, int w)
{
  PyObject *p = (PyObject *) d;
  PyObject *args;

  args = Py_BuildValue("(s,i,i,i)", s, c, v, w);

  if (! args)
    return 0;

  if (!PyEval_CallObject(p, args))
    return 0;
  else
    return 1;
}

static PyObject *
PyMidiIn_New(PyObject *self, PyObject *args)
{
  PyObject *o;
  PyMidiIn *s;
  struct MidiIn *m;

  if (!PyArg_ParseTuple(args, "O;MidiIn.MidiIn", &o))
    return NULL;

  if (!PyCallable_Check(o))
    {
      PyErr_SetString(PyExc_TypeError, "first arg must be callable");
      return 0;
    }

  m = MidiIn_New();
  
  if (!m)
    /* should announce something */
    return 0;

  s = PyObject_NEW(PyMidiIn, &PyMidiIn_Type);
  s->o = m;
  m->cb = callcb;
  m->cb_data = o;

  Py_INCREF(o);

  return (PyObject *) s;
}

static PyMethodDef midiinMethods[] =
{
  {"MidiIn", PyMidiIn_New, METH_VARARGS},
  {0, 0, 0}
};

void
initMidiIn()
{
  Py_InitModule("MidiIn", midiinMethods);
}
#endif

/*#define TEST*/
#ifdef TEST

#include <stdio.h>

int
foo(void *data, const char *s, int c, int u, int v)
{
  printf("(%s %d %d %d)\n", s, c, u, v);
  return 1;
}

int
main()
{
  int c;
  int n = 0;
  struct MidiIn *m = MidiIn_New();

  if (!m)
    return 1;

  m->cb = foo;
  while (n++ < 1000)
    MidiIn_InByte(m, getchar());

  return 0;
}
#endif
